@@ -411,7 +411,7 @@ static int nilfs_btree_node_lookup(const struct nilfs_btree *btree,
 	BUG_ON(indexp == NULL);
 	*indexp = index;
 
-	return (s == 0);
+	return s == 0;
 }
 
 static inline struct nilfs_btree_node *
@@ -1722,9 +1722,8 @@ static int nilfs_btree_propagate_p(struct nilfs_btree *btree,
 				   struct buffer_head *bh)
 {
 	while ((++level < nilfs_btree_height(btree) - 1) &&
-	       !buffer_dirty(path[level].bp_bh) &&
-	       !buffer_prepare_dirty(path[level].bp_bh))
-		nilfs_btnode_mark_prepare_dirty(path[level].bp_bh);
+	       !buffer_dirty(path[level].bp_bh))
+		nilfs_btnode_mark_dirty(path[level].bp_bh);
 
 	return 0;
 }
@@ -1815,8 +1814,7 @@ static int nilfs_btree_prepare_propagate_v(struct nilfs_btree *btree,
 			return ret;
 	}
 	while ((++level < nilfs_btree_height(btree) - 1) &&
-	       !buffer_dirty(path[level].bp_bh) &&
-	       !buffer_prepare_dirty(path[level].bp_bh)) {
+	       !buffer_dirty(path[level].bp_bh)) {
 
 		BUG_ON(buffer_nilfs_volatile(path[level].bp_bh));
 		ret = nilfs_btree_prepare_update_v(btree, path, level);
@@ -1859,12 +1857,24 @@ static int nilfs_btree_propagate_v(struct nilfs_btree *btree,
 				   struct buffer_head *bh)
 {
 	int maxlevel, ret;
+	struct nilfs_btree_node *parent;
+	__u64 ptr;
 
 	get_bh(bh);
 	path[level].bp_bh = bh;
 	ret = nilfs_btree_prepare_propagate_v(btree, path, level, &maxlevel);
 	if (ret < 0)
 		goto out;
+
+	if (buffer_nilfs_volatile(path[level].bp_bh)) {
+		parent = nilfs_btree_get_node(btree, path, level + 1);
+		ptr = nilfs_btree_node_get_ptr(btree, parent,
+					       path[level + 1].bp_index);
+		ret = nilfs_bmap_mark_dirty(&btree->bt_bmap, ptr);
+		if (ret < 0)
+			goto out;
+	}
+
 	nilfs_btree_commit_propagate_v(btree, path, level, maxlevel, bh);
 
  out:
@@ -2008,8 +2018,6 @@ static void nilfs_btree_lookup_dirty_buffers(struct nilfs_bmap *bmap,
 	nilfs_btnode_read_lock(btcache);
 	nilfs_btree_lookup_dirty_buffers_tag(btree, btcache, lists,
 					     PAGECACHE_TAG_DIRTY);
-	nilfs_btree_lookup_dirty_buffers_tag(btree, btcache, lists,
-					     NILFS_PAGECACHE_TAG_PDIRTY);
 	nilfs_btnode_read_unlock(btcache);
 
 	for (level = NILFS_BTREE_LEVEL_NODE_MIN;
