@@ -28,6 +28,7 @@
 #include <linux/proc_fs.h>
 #include <linux/string.h>
 #include <linux/parser.h>
+#include "kern_feature.h"
 #include "nilfs.h"
 #include "sufile.h"
 #include "page.h"
@@ -36,13 +37,6 @@
 static int proc_calc_metrics(char *page, char **start, off_t off,
 			     int count, int *eof, int len);
 
-/*
- * Counters
- */
-void nilfs_init_counters(void)
-{
-}
-
 /*
  * debug info
  */
@@ -84,7 +78,7 @@ static match_table_t class_tokens = {
 
 static const char *find_token(int token, match_table_t tokens)
 {
-	struct match_token *pt;
+	const struct match_token *pt;
 
 	for (pt = tokens; pt->pattern != NULL; pt++)
 		if (pt->token == token)
@@ -177,10 +171,18 @@ static int nilfs_parse_debug_option(char *data)
 
 		token = match_token(++p, opt_tokens, args);
 		switch (token) {
-		case Opt_verbose:  level = 1; break;
-		case Opt_verbose2: level = 2; break;
-		case Opt_verbose3: level = 3; break;
-		case Opt_quiet: level = 0; break;
+		case Opt_verbose:
+			level = 1;
+			break;
+		case Opt_verbose2:
+			level = 2;
+			break;
+		case Opt_verbose3:
+			level = 3;
+			break;
+		case Opt_quiet:
+			level = 0;
+			break;
 		default:
 			goto bad_option;
 		}
@@ -324,8 +326,6 @@ static int snprint_bh_state(char *b, int size, struct buffer_head *bh)
 	TEST_BH_STATE(bh, eopnotsupp, ENOTSUPP, b, size, n, len);
 
 	/* nilfs private */
-	TEST_BH_STATE(bh, prepare_dirty, Prepare_Dirty, b, size, n, len);
-	TEST_BH_STATE(bh, nilfs_freeze, Freeze, b, size, n, len);
 	TEST_BH_STATE(bh, nilfs_allocated, Allocated, b, size, n, len);
 	TEST_BH_STATE(bh, nilfs_node, Node, b, size, n, len);
 	TEST_BH_STATE(bh, nilfs_volatile, Volatile, b, size, n, len);
@@ -655,7 +655,7 @@ int nilfs_init_proc_entries(void)
 {
 	struct proc_dir_entry *entry;
 
-	nilfs_proc_root = proc_mkdir("nilfs2", proc_root_fs);
+	nilfs_proc_root = proc_mkdir("fs/nilfs2", NULL);
 	if (!nilfs_proc_root) {
 		printk(KERN_WARNING "NILFS: cannot create proc root\n");
 		return 0; /* We don't abort when failed to make proc entries */
@@ -681,7 +681,7 @@ void nilfs_remove_proc_entries(void)
 {
 	remove_proc_entry("page", nilfs_proc_root);
 	remove_proc_entry("debug_option", nilfs_proc_root);
-	remove_proc_entry("nilfs2", proc_root_fs);
+	remove_proc_entry("fs/nilfs2", NULL);
 }
 
 /*
@@ -700,10 +700,10 @@ int nilfs_releasepage(struct page *page, gfp_t gfp_mask)
 			verbose = 1;
 	}
 	if (unlikely(!PagePrivate(page)))
-		PAGE_BUG(page, "no buffers");
+		NILFS_PAGE_BUG(page, "no buffers");
 
 	if (buffer_nilfs_allocated(page_buffers(page)))
-		PAGE_BUG(page, "nilfs allocated page");
+		NILFS_PAGE_BUG(page, "nilfs allocated page");
 
 	/*
 	 * Note that non-busy buffer heads may be discarded though the
@@ -759,11 +759,10 @@ nilfs_invalidatepage(struct page *page, unsigned long offset)
 /*
  * Radix-tree checker
  */
-#define S_N_PAGEVEC  16
 void nilfs_check_radix_tree(const char *fname, int line,
 			    struct address_space *mapping, int tag)
 {
-	struct page *pages[S_N_PAGEVEC];
+	struct pagevec pvec;
 	unsigned int i, n;
 	pgoff_t index = 0;
 	char *page_type;
@@ -776,14 +775,15 @@ void nilfs_check_radix_tree(const char *fname, int line,
 	else
 		page_type = "leaking";
 
+	pagevec_init(&pvec, 0);
  repeat:
 	if (tag < 0) {
-		n = find_get_pages(mapping, index, S_N_PAGEVEC, pages);
+		n = pagevec_lookup(&pvec, mapping, index, PAGEVEC_SIZE);
 		if (n)
-			index = pages[n - 1]->index + 1;
+			index = pvec.pages[n - 1]->index + 1;
 	} else
-		n = find_get_pages_tag(mapping, &index, tag, S_N_PAGEVEC,
-				       pages);
+		n = pagevec_lookup_tag(&pvec, mapping, &index, tag,
+				       PAGEVEC_SIZE);
 	if (!n) {
 		if (nr_found)
 			printk(KERN_WARNING "%s: found %d %s pages\n",
@@ -792,16 +792,19 @@ void nilfs_check_radix_tree(const char *fname, int line,
 	}
 
 	for (i = 0; i < n; i++) {
-		nilfs_page_debug(fname, line, pages[i], "%s page", page_type);
+		nilfs_page_debug(fname, line, pvec.pages[i], "%s page",
+				 page_type);
 		nr_found++;
-		page_cache_release(pages[i]);
 	}
+	pagevec_release(&pvec);
+	cond_resched();
 	goto repeat;
 }
 
 /*
  * btnode cache checker
  */
+#define S_N_PAGEVEC  16
 void nilfs_check_btnode_cache(const char *fname, int line,
 			      struct nilfs_btnode_cache *btnc, int tag)
 {
@@ -813,8 +816,6 @@ void nilfs_check_btnode_cache(const char *fname, int line,
 
 	if (tag == PAGECACHE_TAG_DIRTY)
 		page_type = "dirty";
-	else if (tag == NILFS_PAGECACHE_TAG_PDIRTY)
-		page_type = "pdirty";
 	else
 		page_type = "leaking";
 
