@@ -171,7 +171,7 @@ int nilfs_mdt_create_block(struct inode *inode, unsigned long block,
 			   nilfs_mdt_init_block_t *init_block)
 {
 	struct the_nilfs *nilfs = NILFS_MDT(inode)->mi_nilfs;
-	struct nilfs_sb_info *sbi = NULL;
+	struct nilfs_sb_info *writer = NULL;
 	struct super_block *sb = inode->i_sb;
 	struct nilfs_transaction_info ti;
 	struct buffer_head *bh;
@@ -180,12 +180,12 @@ int nilfs_mdt_create_block(struct inode *inode, unsigned long block,
 	mdt_debug(3, "called (ino=%lu, blkoff=%lu)\n", inode->i_ino, block);
 
 	if (!sb) {
-		sbi = nilfs_get_writer(nilfs);
-		if (!sbi) {
+		writer = nilfs_get_writer(nilfs);
+		if (!writer) {
 			err = -EROFS;
 			goto out;
 		}
-		sb = sbi->s_super;
+		sb = writer->s_super;
 	}
 
 	nilfs_transaction_begin(sb, &ti, 0);
@@ -218,7 +218,7 @@ int nilfs_mdt_create_block(struct inode *inode, unsigned long block,
 
  failed_unlock:
 	nilfs_transaction_end(sb, !err);
-	if (sbi)
+	if (writer)
 		nilfs_put_writer(nilfs);
  out:
 	mdt_debug(3, "done (err=%d)\n", err);
@@ -435,39 +435,6 @@ int nilfs_mdt_delete_block(struct inode *inode, unsigned long block)
 	return err;
 }
 
-/**
- * nilfs_mdt_truncate_blocks - truncate meta data file.
- * @inode: inode of the meta data file
- * @block: block offset
- *
- * Return Value: On success, zero is returned.
- * On error, one of the following negative error code is returned.
- *
- * %-ENOMEM - Insufficient memory available.
- *
- * %-EIO - I/O error
- *
- * %-EINVAL - bmap is broken. (the caller should call nilfs_error())
- */
-int nilfs_mdt_truncate_blocks(struct inode *inode, unsigned long block)
-{
-	struct nilfs_inode_info *ii = NILFS_I(inode);
-	int err;
-
-	mdt_debug(3, "called (ino=%lu, blkoff=%lu)\n", inode->i_ino, block);
-	err = nilfs_bmap_truncate(ii->i_bmap, block);
-	if (likely(!err)) {
-#if 0  /* XXX: truncation of page cache should be delayed to avoid excessive
-	  page removals and insertions */
-		truncate_inode_pages(inode->i_mapping,
-				     (loff_t)block << inode->i_blkbits);
-#endif
-		nilfs_mdt_mark_dirty(inode);
-	}
-	mdt_debug(3, "done (err=%d)\n", err);
-	return err;
-}
-
 /**
  * nilfs_mdt_forget_block - discard dirty state and try to remove the page
  * @inode: inode of the meta data file
@@ -577,7 +544,9 @@ nilfs_mdt_write_page(struct page *page, struct writeback_control *wbc)
 {
 	struct inode *inode = container_of(page->mapping,
 					   struct inode, i_data);
-	int err;
+	struct super_block *sb = inode->i_sb;
+	struct nilfs_sb_info *writer = NULL;
+	int err = 0;
 
 	mdt_debug(2, "called (page=%p, index=%lu, wbc nonblocking %d, "
 		  "wbc for_reclaim %d)\n",
@@ -585,15 +554,23 @@ nilfs_mdt_write_page(struct page *page, struct writeback_control *wbc)
 	redirty_page_for_writepage(wbc, page);
 	unlock_page(page);
 
-	if (!inode->i_sb)
-		return 0;
-	if (wbc->sync_mode == WB_SYNC_ALL) {
-		err = nilfs_construct_segment(inode->i_sb);
-		if (unlikely(err))
-			return err;
-	} else if (wbc->for_reclaim)
-		nilfs_flush_segment(NILFS_SB(inode->i_sb), inode->i_ino);
-	return 0;
+	if (NILFS_MDT(inode)->mi_orig_inode)
+		return 0; /* Do not request flush for shadow inode */
+	if (!sb) {
+		writer = nilfs_get_writer(NILFS_MDT(inode)->mi_nilfs);
+		if (!writer)
+			return -EROFS;
+		sb = writer->s_super;
+	}
+
+	if (wbc->sync_mode == WB_SYNC_ALL)
+		err = nilfs_construct_segment(sb);
+	else if (wbc->for_reclaim)
+		nilfs_flush_segment(sb, inode->i_ino);
+
+	if (writer)
+		nilfs_put_writer(NILFS_MDT(inode)->mi_nilfs);
+	return err;
 }
 
 
