@@ -164,10 +164,10 @@ static void nilfs_free_page(struct page *page)
 #ifdef CONFIG_NILFS_DEBUG
 	/* Checks helpful for debugging purpose. */
 	if (unlikely(!list_empty(&page->lru)))
-		PAGE_BUG(page, "page not isolated");
+		NILFS_PAGE_BUG(page, "page not isolated");
 
 	if (unlikely(page_count(page) != 1))
-		PAGE_BUG(page, "wrong page count");
+		NILFS_PAGE_BUG(page, "wrong page count");
 
 	if (unlikely(PageDirty(page)))
 		PAGE_DEBUG(page, "dirty page");
@@ -435,12 +435,12 @@ void nilfs_free_buffer_page(struct page *page)
 {
 	BUG_ON(!PageLocked(page));
 	if (page->mapping)
-		PAGE_BUG(page, "freeing page with mapping");
+		NILFS_PAGE_BUG(page, "freeing page with mapping");
 
 	if (unlikely(!page_has_buffers(page)))
 		PAGE_DEBUG(page, "freeing page without buffers");
 	else if (!try_to_free_buffers(page))
-		PAGE_BUG(page, "failed to free page");
+		NILFS_PAGE_BUG(page, "failed to free page");
 
 	nilfs_free_page(page);
 }
@@ -539,7 +539,7 @@ void nilfs_copy_buffer_page(struct page *src, struct page *dst, int copy_dirty)
 	}
 
 	if (copy_dirty)
-		mask |= (1UL << BH_Dirty) | (1UL << BH_Prepare_Dirty);
+		mask |= (1UL << BH_Dirty);
 
 	dbh = dbufs = page_buffers(dst);
 	do {
@@ -569,32 +569,18 @@ void nilfs_copy_buffer_page(struct page *src, struct page *dst, int copy_dirty)
  *
  * nilfs_page_buffers_clean() returns zero if the page has dirty buffers.
  * Otherwise, it returns non-zero value.
- * For the btnode page, the prepare-dirty state flag for written buffers
- * cleared here, and return values are:
- *	00(b): page state unchanged (remains dirty or prepare-dirty)
- *	01(b): page state will be changed from dirty to clean
- *	10(b): page state will be changed from prepare-dirty to dirty
- *	11(b): page state will be changed from prepare-dirty to clean
  */
 int nilfs_page_buffers_clean(struct page *page)
 {
 	struct buffer_head *bh, *head;
-	int d = 0, pd = 0, pc = 0;
 
 	bh = head = page_buffers(page);
 	do {
-		int _d = buffer_dirty(bh), _p = buffer_prepare_dirty(bh);
-		d |= _d;
-		pd |= _p & _d;
-		if (_p & !_d) {
-			/* Note: buffer's prepare_dirty bit is cleared here */
-			clear_buffer_prepare_dirty(bh);
-			pc = 1;
-		}
+		if (buffer_dirty(bh))
+			return 0;
 		bh = bh->b_this_page;
 	} while (bh != head);
-	return (!pd & pc) << NILFS_PAGECACHE_TAG_PDIRTY |
-		!d << PAGECACHE_TAG_DIRTY;
+	return 1;
 }
 
 unsigned nilfs_page_count_clean_buffers(struct page *page,
@@ -620,8 +606,8 @@ void nilfs_page_bug(struct page *page)
 	unsigned long ino = 0;
 
 	if (unlikely(!page)) {
-		printk(KERN_CRIT "PAGE_BUG(NULL)\n");
-		BUG();
+		printk(KERN_CRIT "NILFS_PAGE_BUG(NULL)\n");
+		return;
 	}
 
 	m = page->mapping;
@@ -630,7 +616,7 @@ void nilfs_page_bug(struct page *page)
 		if (inode != NULL)
 			ino = inode->i_ino;
 	}
-	printk(KERN_CRIT "PAGE_BUG(%p): cnt=%d index#=%llu flags=0x%lx "
+	printk(KERN_CRIT "NILFS_PAGE_BUG(%p): cnt=%d index#=%llu flags=0x%lx "
 	       "mapping=%p ino=%lu\n",
 	       page, atomic_read(&page->_count),
 	       (unsigned long long)page->index, page->flags, m, ino);
@@ -648,7 +634,6 @@ void nilfs_page_bug(struct page *page)
 			bh = bh->b_this_page;
 		} while (bh != head);
 	}
-	BUG();
 }
 
 #if !HAVE_CLEAR_PAGE_DIRTY
