@@ -56,14 +56,13 @@
 
 /* Construction mode */
 enum {
-	SC_FLUSH_DATA = 1,   /* Flush current dirty data blocks and make
-				partial segments without the super root and
-				the inode file */
-	SC_FLUSH_IFILE,      /* Flush current dirty data blocks and ifile;
-				make partial segments without checkpoint */
-	SC_LSEG_SR,          /* Make a logical segment having a super root */
-	SC_LSEG_DSYNC,       /* Flush data blocks of a given file and make
-				a logical segment without the super root */
+	SC_LSEG_SR = 1,	/* Make a logical segment having a super root */
+	SC_LSEG_DSYNC,	/* Flush data blocks of a given file and make
+			   a logical segment without a super root */
+	SC_FLUSH_FILE,	/* Flush data files, leads to segment writes without
+			   creating a checkpoint */
+	SC_FLUSH_DAT,	/* Flush DAT file. This also creates segments without
+			   a checkpoint */
 };
 
 /*
@@ -79,8 +78,8 @@ enum {
 	SC_MAIN_SUFILE,
 	SC_MAIN_DAT,
 	SC_MAIN_SR,
-	SC_MAIN_DONE,
 	SC_MAIN_DSYNC,
+	SC_MAIN_DONE,
 };
 
 enum {
@@ -136,7 +135,8 @@ struct nilfs_sc_operations {
  * Other definitions
  */
 static void nilfs_segctor_start_timer(struct nilfs_sc_info *);
-static void nilfs_segctor_do_flush(struct nilfs_sc_info *, unsigned long);
+static void nilfs_segctor_do_flush(struct nilfs_sc_info *, int);
+static void nilfs_segctor_do_immediate_flush(struct nilfs_sc_info *);
 static void nilfs_dispose_list(struct nilfs_sb_info *, struct list_head *,
 			       int);
 
@@ -180,7 +180,7 @@ int nilfs_init_transaction_cache(void)
 #else
 				  NULL);
 #endif
-	return ((nilfs_transaction_cachep == NULL) ? -ENOMEM : 0);
+	return (nilfs_transaction_cachep == NULL) ? -ENOMEM : 0;
 }
 
 /**
@@ -330,7 +330,7 @@ int nilfs_transaction_end(struct super_block *sb, int commit)
 			nilfs_segctor_start_timer(sci);
 		if (atomic_read(&sbi->s_nilfs->ns_ndirtyblks) >
 		    sci->sc_watermark)
-			nilfs_segctor_do_flush(sci, NILFS_SEGCTOR_FLUSH_DATA);
+			nilfs_segctor_do_flush(sci, 0);
 	}
 	up_read(&sbi->s_nilfs->ns_segctor_sem);
 	seg_debug(3, "task %p unlocked segment semaphore\n", current);
@@ -343,6 +343,30 @@ int nilfs_transaction_end(struct super_block *sb, int commit)
 	return err;
 }
 
+void nilfs_relax_pressure_in_lock(struct super_block *sb)
+{
+	struct nilfs_sb_info *sbi = NILFS_SB(sb);
+	struct nilfs_sc_info *sci = NILFS_SC(sbi);
+	struct the_nilfs *nilfs = sbi->s_nilfs;
+
+	if (!sci || !sci->sc_flush_request)
+		return;
+
+	set_bit(NILFS_SC_PRIOR_FLUSH, &sci->sc_flags);
+	up_read(&nilfs->ns_segctor_sem);
+
+	down_write(&nilfs->ns_segctor_sem);
+	if (sci->sc_flush_request &&
+	    test_bit(NILFS_SC_PRIOR_FLUSH, &sci->sc_flags)) {
+		struct nilfs_transaction_info *ti = current->journal_info;
+
+		ti->ti_flags |= NILFS_TI_WRITER;
+		nilfs_segctor_do_immediate_flush(sci);
+		ti->ti_flags &= ~NILFS_TI_WRITER;
+	}
+	downgrade_write(&nilfs->ns_segctor_sem);
+}
+
 static void nilfs_transaction_lock(struct nilfs_sb_info *sbi,
 				   struct nilfs_transaction_info *ti,
 				   int gcflag)
@@ -352,8 +376,6 @@ static void nilfs_transaction_lock(struct nilfs_sb_info *sbi,
 	BUG_ON(cur_ti);
 	BUG_ON(!ti);
 	ti->ti_flags = NILFS_TI_WRITER;
-	if (gcflag)
-		ti->ti_flags |= NILFS_TI_GC;
 	ti->ti_count = 0;
 	ti->ti_save = cur_ti;
 	ti->ti_magic = NILFS_TI_MAGIC;
@@ -361,7 +383,18 @@ static void nilfs_transaction_lock(struct nilfs_sb_info *sbi,
 	current->journal_info = ti;
 
 	seg_debug(3, "task %p locking segment semaphore\n", current);
-	down_write(&sbi->s_nilfs->ns_segctor_sem);
+	for (;;) {
+		down_write(&sbi->s_nilfs->ns_segctor_sem);
+		if (!test_bit(NILFS_SC_PRIOR_FLUSH, &NILFS_SC(sbi)->sc_flags))
+			break;
+
+		nilfs_segctor_do_immediate_flush(NILFS_SC(sbi));
+
+		up_write(&sbi->s_nilfs->ns_segctor_sem);
+		yield();
+	}
+	if (gcflag)
+		ti->ti_flags |= NILFS_TI_GC;
 	seg_debug(3, "locked\n");
 }
 
@@ -582,7 +615,7 @@ static int nilfs_collect_file_node(struct nilfs_sc_info *sci,
 {
 	int err;
 
-	/* BUG_ON(!nilfs_btnode_buffer_dirty(bh)); */
+	/* BUG_ON(!buffer_dirty(bh)); */
 	/* excluded by scan_dirty_node_buffers() */
 	err = nilfs_bmap_propagate(NILFS_I(inode)->i_bmap, bh);
 	if (unlikely(err < 0))
@@ -794,94 +827,68 @@ static int nilfs_prepare_data_page(struct inode *inode, struct page *page)
 	return err;
 }
 
-static int
-nilfs_segctor_scan_dirty_data_buffers(struct nilfs_sc_info *sci,
-				      struct inode *inode,
-				      int (*collect)(struct nilfs_sc_info *,
-						     struct buffer_head *,
-						     struct inode *))
+static int nilfs_lookup_dirty_data_buffers(struct inode *inode,
+					   struct list_head *listp,
+					   struct nilfs_sc_info *sci)
 {
+	struct nilfs_segment_buffer *segbuf = sci->sc_curseg;
 	struct address_space *mapping = inode->i_mapping;
-	struct page *pages[SC_N_PAGEVEC];
-	unsigned int i, n, ndirties;
+	struct pagevec pvec;
+	unsigned i, ndirties = 0, nlimit;
 	pgoff_t index = 0;
-	int err = 0, can_copy = !S_ISDIR(inode->i_mode);
+	int err = 0;
 
 	seg_debug(3, "called (ino=%lu)\n", inode->i_ino);
+	nlimit = sci->sc_segbuf_nblocks -
+		(sci->sc_nblk_this_inc + segbuf->sb_sum.nblocks);
+	pagevec_init(&pvec, 0);
  repeat:
-	n = find_get_pages_tag(mapping, &index, PAGECACHE_TAG_DIRTY,
-			       SC_N_PAGEVEC, pages);
-	if (!n) {
+	if (!pagevec_lookup_tag(&pvec, mapping, &index, PAGECACHE_TAG_DIRTY,
+				PAGEVEC_SIZE)) {
 		seg_debug(3, "done (ino=%lu)\n", inode->i_ino);
 		return 0;
 	}
-	for (i = 0; i < n; i++) {
+	for (i = 0; i < pagevec_count(&pvec); i++) {
 		struct buffer_head *bh, *head;
-		struct page *page = pages[i];
-
-		if (err)
-			goto skip_page;
+		struct page *page = pvec.pages[i];
 
 		if (mapping->host) {
 			err = nilfs_prepare_data_page(inode, page);
 			if (unlikely(err))
-				goto skip_page;
+				break;
 		}
 
 		bh = head = page_buffers(page);
-		ndirties = 0;
 		do {
 			if (buffer_dirty(bh)) {
-				get_bh(bh);
-				err = collect(sci, bh, inode);
-				put_bh(bh);
-				if (unlikely(err)) {
-					if (!ndirties || err != -E2BIG)
-						goto skip_page;
+				if (ndirties > nlimit) {
+					err = -E2BIG;
 					break;
-					/* each blocks in a mmapped
-					   page should be copied */
 				}
+				get_bh(bh);
+				list_add_tail(&bh->b_assoc_buffers, listp);
 				ndirties++;
 			}
 			bh = bh->b_this_page;
 		} while (bh != head);
-
-		if (can_copy) {
-			/* Decide whether to copy buffer or not.
-			   This must be done after a writeback flag
-			   is set on the page */
-			if (page_mapped(page)) {
-				nilfs_set_page_to_be_frozen(page);
-				ClearPageChecked(page);
-			} else if (PageChecked(page)) {
-				nilfs_set_page_to_be_frozen(page);
-			}
-		}
- skip_page:
-		page_cache_release(page);
 	}
+	pagevec_release(&pvec);
+	cond_resched();
+
 	if (!err)
 		goto repeat;
-
 	seg_debug(3, "failed (err=%d, ino=%lu)\n", err, inode->i_ino);
 	return err;
 }
 
-static int
-nilfs_segctor_scan_dirty_node_buffers(struct nilfs_sc_info *sci,
-				      struct inode *inode,
-				      int (*collect)(struct nilfs_sc_info *,
-						     struct buffer_head *,
-						     struct inode *))
+static void nilfs_lookup_dirty_node_buffers(struct inode *inode,
+					    struct list_head *listp)
 {
 	struct nilfs_inode_info *ii = NILFS_I(inode);
 	struct page *pages[SC_N_PAGEVEC];
 	struct buffer_head *bh, *head;
 	unsigned int i, n;
 	pgoff_t index = 0;
-	LIST_HEAD(node_buffers);
-	int err = 0;
 
 	seg_debug(3, "called (ino=%lu)\n", inode->i_ino);
 
@@ -890,61 +897,21 @@ nilfs_segctor_scan_dirty_node_buffers(struct nilfs_sc_info *sci,
 					    pages, &index, SC_N_PAGEVEC,
 					    PAGECACHE_TAG_DIRTY);
 	if (!n)
-		goto end_lookup;
+		return;
 
 	for (i = 0; i < n; i++) {
 		bh = head = page_buffers(pages[i]);
 		do {
-			if (nilfs_btnode_buffer_dirty(bh)) {
+			if (buffer_dirty(bh)) {
 				get_bh(bh);
-				list_add_tail(&bh->b_assoc_buffers,
-					      &node_buffers);
+				list_add_tail(&bh->b_assoc_buffers, listp);
 			}
-		} while ((bh = bh->b_this_page) != head);
+			bh = bh->b_this_page;
+		} while (bh != head);
+
 		page_cache_release(pages[i]);
 	}
 	goto repeat;
-
- end_lookup:
-	list_for_each_entry_safe(bh, head, &node_buffers, b_assoc_buffers) {
-		list_del_init(&bh->b_assoc_buffers);
-		if (likely(!err))
-			err = collect(sci, bh, inode);
-		brelse(bh);
-	}
-	seg_debug(3, "done (err=%d, ino=%lu)\n", err, inode->i_ino);
-	return err;
-}
-
-static int
-nilfs_segctor_scan_dirty_bmap_buffers(struct nilfs_sc_info *sci,
-				      struct inode *inode,
-				      int (*collect)(struct nilfs_sc_info *,
-						     struct buffer_head *,
-						     struct inode *))
-{
-	struct nilfs_inode_info *ii = NILFS_I(inode);
-	struct buffer_head *bh, *n;
-	LIST_HEAD(node_buffers);
-	int err = 0;
-
-	nilfs_bmap_lookup_dirty_buffers(ii->i_bmap, &node_buffers);
-	list_for_each_entry_safe(bh, n, &node_buffers, b_assoc_buffers) {
-		list_del_init(&bh->b_assoc_buffers);
-		err = collect(sci, bh, inode);
-		brelse(bh);
-		if (unlikely(err)) {
-			while (!list_empty(&node_buffers)) {
-				bh = list_entry(node_buffers.next,
-						struct buffer_head,
-						b_assoc_buffers);
-				list_del_init(&bh->b_assoc_buffers);
-				brelse(bh);
-			}
-			break;
-		}
-	}
-	return err;
 }
 
 static void nilfs_dispose_list(struct nilfs_sb_info *sbi,
@@ -1001,11 +968,11 @@ static int nilfs_test_metadata_dirty(struct nilfs_sb_info *sbi)
 
 static int nilfs_segctor_clean(struct nilfs_sc_info *sci)
 {
-	return (list_empty(&sci->sc_dirty_files) &&
+	return list_empty(&sci->sc_dirty_files) &&
 		!test_bit(NILFS_SC_DIRTY, &sci->sc_flags) &&
 		list_empty(&sci->sc_cleaning_segments) &&
 		(!test_bit(NILFS_SC_GC_COPY, &sci->sc_flags) ||
-		 list_empty(&sci->sc_gc_inodes)));
+		 list_empty(&sci->sc_gc_inodes));
 }
 
 static int nilfs_segctor_confirm(struct nilfs_sc_info *sci)
@@ -1025,28 +992,11 @@ static int nilfs_segctor_confirm(struct nilfs_sc_info *sci)
 	return ret;
 }
 
-static int nilfs_segctor_reconfirm(struct nilfs_sc_info *sci)
-{
-	if (nilfs_test_metadata_dirty(sci->sc_sbi))
-		set_bit(NILFS_SC_DIRTY, &sci->sc_flags);
-
-	if (nilfs_segctor_clean(sci)) {
-		seg_debug(2, "Aborted construction (no changes found in "
-			  "reconfirmation)\n");
-		return 1;
-	}
-	return 0;
-}
-
-static void
-nilfs_segctor_clear_metadata_dirty(struct nilfs_sc_info *sci, int mode)
+static void nilfs_segctor_clear_metadata_dirty(struct nilfs_sc_info *sci)
 {
 	struct nilfs_sb_info *sbi = sci->sc_sbi;
 	struct the_nilfs *nilfs = sbi->s_nilfs;
 
-	if (mode == SC_LSEG_DSYNC)
-		return;
-
 	if (SC_STAGE_DONE(&sci->sc_stage, SC_MAIN_IFILE))
 		nilfs_mdt_clear_dirty(sbi->s_ifile);
 	if (SC_STAGE_DONE(&sci->sc_stage, SC_MAIN_CPFILE))
@@ -1282,31 +1232,82 @@ static void nilfs_segctor_commit_free_segments(struct nilfs_sc_info *sci)
 	nilfs_dispose_segment_list(&sci->sc_cleaning_segments);
 }
 
+static int nilfs_segctor_apply_buffers(struct nilfs_sc_info *sci,
+				       struct inode *inode,
+				       struct list_head *listp,
+				       int (*collect)(struct nilfs_sc_info *,
+						      struct buffer_head *,
+						      struct inode *))
+{
+	struct buffer_head *bh, *n;
+	int err = 0;
+
+	if (collect) {
+		list_for_each_entry_safe(bh, n, listp, b_assoc_buffers) {
+			list_del_init(&bh->b_assoc_buffers);
+			err = collect(sci, bh, inode);
+			brelse(bh);
+			if (unlikely(err))
+				goto dispose_buffers;
+		}
+		return 0;
+	}
+
+ dispose_buffers:
+	while (!list_empty(listp)) {
+		bh = list_entry(listp->next, struct buffer_head,
+				b_assoc_buffers);
+		list_del_init(&bh->b_assoc_buffers);
+		brelse(bh);
+	}
+	return err;
+}
+
 static int nilfs_segctor_scan_file(struct nilfs_sc_info *sci,
 				   struct inode *inode,
 				   struct nilfs_sc_operations *sc_ops)
 {
-	int err = 0;
+	LIST_HEAD(data_buffers);
+	LIST_HEAD(node_buffers);
+	int err, err2;
 
 	seg_debug(3, "called (ino=%lu, main_stage=%d)\n",
 		  inode->i_ino, sci->sc_stage.main);
 
 	if (sci->sc_stage.sub == SC_SUB_DATA) {
-		err = nilfs_segctor_scan_dirty_data_buffers(
-			sci, inode, sc_ops->collect_data);
-		if (unlikely(err))
+		err = nilfs_lookup_dirty_data_buffers(inode, &data_buffers,
+						      sci);
+		if (err) {
+			err2 = nilfs_segctor_apply_buffers(
+				sci, inode, &data_buffers,
+				err == -E2BIG ? sc_ops->collect_data : NULL);
+			if (err == -E2BIG)
+				err = err2;
 			goto break_or_fail;
+		}
+	}
+	nilfs_lookup_dirty_node_buffers(inode, &node_buffers);
 
+	if (sci->sc_stage.sub == SC_SUB_DATA) {
+		err = nilfs_segctor_apply_buffers(
+			sci, inode, &data_buffers, sc_ops->collect_data);
+		if (unlikely(err)) {
+			/* dispose node list */
+			nilfs_segctor_apply_buffers(
+				sci, inode, &node_buffers, NULL);
+			goto break_or_fail;
+		}
 		sci->sc_stage.sub++;
 	}
-	/* sci->sc_stage.sub == SC_SUB_NODE */
-	err = nilfs_segctor_scan_dirty_node_buffers(
-		sci, inode, sc_ops->collect_node);
+	/* sci->sc_state.sub == SC_SUB_NODE */
+	err = nilfs_segctor_apply_buffers(
+		sci, inode, &node_buffers, sc_ops->collect_node);
 	if (unlikely(err))
 		goto break_or_fail;
 
-	err = nilfs_segctor_scan_dirty_bmap_buffers(
-		sci, inode, sc_ops->collect_bmap);
+	nilfs_bmap_lookup_dirty_buffers(NILFS_I(inode)->i_bmap, &node_buffers);
+	err = nilfs_segctor_apply_buffers(
+		sci, inode, &node_buffers, sc_ops->collect_bmap);
 	if (unlikely(err))
 		goto break_or_fail;
 
@@ -1318,6 +1319,23 @@ static int nilfs_segctor_scan_file(struct nilfs_sc_info *sci,
 	return err;
 }
 
+static int nilfs_segctor_scan_file_dsync(struct nilfs_sc_info *sci,
+					 struct inode *inode)
+{
+	LIST_HEAD(data_buffers);
+	int err, err2;
+
+	err = nilfs_lookup_dirty_data_buffers(inode, &data_buffers, sci);
+	err2 = nilfs_segctor_apply_buffers(sci, inode, &data_buffers,
+					   (!err || err == -E2BIG) ?
+					   nilfs_collect_file_data : NULL);
+	if (err == -E2BIG)
+		err = err2;
+	if (!err)
+		nilfs_segctor_end_finfo(sci, inode);
+	return err;
+}
+
 static int nilfs_segctor_collect_blocks(struct nilfs_sc_info *sci, int mode)
 {
 	struct nilfs_sb_info *sbi = sci->sc_sbi;
@@ -1326,7 +1344,6 @@ static int nilfs_segctor_collect_blocks(struct nilfs_sc_info *sci, int mode)
 	struct nilfs_inode_info *ii;
 	int err = 0;
 
- start:
 	switch (sci->sc_stage.main) {
 	case SC_MAIN_INIT:
 		/*
@@ -1340,7 +1357,7 @@ static int nilfs_segctor_collect_blocks(struct nilfs_sc_info *sci, int mode)
 				seg_debug(2, "** DSYNC BEGIN\n");
 				SC_STAGE_SKIP_TO(&sci->sc_stage,
 						 SC_MAIN_DSYNC);
-				goto start;
+				goto dsync_mode;
 			}
 			seg_debug(2, "** LSEG BEGIN\n");
 		} else
@@ -1348,6 +1365,10 @@ static int nilfs_segctor_collect_blocks(struct nilfs_sc_info *sci, int mode)
 
 		sci->sc_stage.dirty_file_ptr = NULL;
 		sci->sc_stage.gc_inode_ptr = NULL;
+		if (mode == SC_FLUSH_DAT) {
+			SC_STAGE_SKIP_TO(&sci->sc_stage, SC_MAIN_DAT);
+			goto dat_stage;
+		}
 		SC_STAGE_NEXT(&sci->sc_stage);
 	case SC_MAIN_GC:
 		seg_debug(3, "** GC INODE STAGE\n");
@@ -1392,7 +1413,7 @@ static int nilfs_segctor_collect_blocks(struct nilfs_sc_info *sci, int mode)
 			/* XXX: required ? */
 		}
 		sci->sc_stage.dirty_file_ptr = NULL;
-		if (mode == SC_FLUSH_DATA) {
+		if (mode == SC_FLUSH_FILE) {
 			SC_STAGE_SKIP_TO(&sci->sc_stage, SC_MAIN_DONE);
 			seg_debug(2, "** LSEG CONTINUED\n");
 			return 0;
@@ -1425,11 +1446,6 @@ static int nilfs_segctor_collect_blocks(struct nilfs_sc_info *sci, int mode)
 					      &nilfs_sc_file_ops);
 		if (unlikely(err))
 			break;
-		if (mode == SC_FLUSH_IFILE) {
-			SC_STAGE_SKIP_TO(&sci->sc_stage, SC_MAIN_DONE);
-			seg_debug(2, "** LSEG CONTINUED\n");
-			return 0;
-		}
 		SC_STAGE_NEXT(&sci->sc_stage);
 		/* Creating a checkpoint */
 		err = nilfs_segctor_create_checkpoint(sci);
@@ -1454,11 +1470,17 @@ static int nilfs_segctor_collect_blocks(struct nilfs_sc_info *sci, int mode)
 			break;
 		SC_STAGE_NEXT(&sci->sc_stage);
 	case SC_MAIN_DAT:
+ dat_stage:
 		seg_debug(3, "** DAT STAGE\n");
 		err = nilfs_segctor_scan_file(sci, nilfs_dat_inode(nilfs),
 					      &nilfs_sc_dat_ops);
 		if (unlikely(err))
 			break;
+		if (mode == SC_FLUSH_DAT) {
+			SC_STAGE_SKIP_TO(&sci->sc_stage, SC_MAIN_DONE);
+			seg_debug(2, "** LSEG CONTINUED\n");
+			return 0;
+		}
 		SC_STAGE_NEXT(&sci->sc_stage);
 	case SC_MAIN_SR:
 		seg_debug(3, "** SR STAGE\n");
@@ -1469,30 +1491,28 @@ static int nilfs_segctor_collect_blocks(struct nilfs_sc_info *sci, int mode)
 				break;
 			seg_debug(3, "add a super root block\n");
 		}
-		SC_STAGE_NEXT(&sci->sc_stage);
-	case SC_MAIN_DONE:
-		/*
-		 * Post processes after final segment construction
-		 * can be inserted here.
-		 */
+		/* End of a logical segment */
 		sci->sc_curseg->sb_sum.flags |= NILFS_SS_LOGEND;
+		SC_STAGE_SKIP_TO(&sci->sc_stage, SC_MAIN_DONE);
 		seg_debug(2, "** LSEG END\n");
 		return 0;
 	case SC_MAIN_DSYNC:
+ dsync_mode:
 		sci->sc_curseg->sb_sum.flags |= NILFS_SS_SYNDT;
 		ii = sci->sc_stage.dirty_file_ptr;
 		if (!test_bit(NILFS_I_BUSY, &ii->i_state))
 			break;
-		err = nilfs_segctor_scan_dirty_data_buffers(
-			sci, &ii->vfs_inode, nilfs_collect_file_data);
+
+		err = nilfs_segctor_scan_file_dsync(sci, &ii->vfs_inode);
 		if (unlikely(err))
 			break;
-		nilfs_segctor_end_finfo(sci, &ii->vfs_inode);
 		sci->sc_stage.dirty_file_ptr = NULL;
 		sci->sc_curseg->sb_sum.flags |= NILFS_SS_LOGEND;
 		SC_STAGE_SKIP_TO(&sci->sc_stage, SC_MAIN_DONE);
 		seg_debug(2, "** DSYNC END\n");
 		return 0;
+	case SC_MAIN_DONE:
+		return 0;
 	default:
 		BUG();
 	}
@@ -1571,6 +1591,7 @@ static int nilfs_segctor_begin_construction(struct nilfs_sc_info *sci,
 		nilfs_shift_to_next_segment(nilfs);
 		err = nilfs_segbuf_map(segbuf, nilfs->ns_segnum, 0, nilfs);
 	}
+	sci->sc_segbuf_nblocks = segbuf->sb_rest_blocks;
 
 	err = nilfs_touch_segusage(sufile, segbuf->sb_segnum);
 	if (unlikely(err))
@@ -1628,6 +1649,8 @@ static int nilfs_segctor_extend_segments(struct nilfs_sc_info *sci,
 		if (unlikely(err))
 			goto failed_segbuf;
 
+		sci->sc_segbuf_nblocks += segbuf->sb_rest_blocks;
+
 		/* allocate the next next full segment */
 		err = nilfs_sufile_alloc(sufile, &nextnextnum);
 		if (unlikely(err))
@@ -1781,6 +1804,7 @@ static void nilfs_segctor_truncate_segments(struct nilfs_sc_info *sci,
 	list_for_each_entry_safe_continue(segbuf, n, &sci->sc_segbufs,
 					  sb_list) {
 		list_del_init(&segbuf->sb_list);
+		sci->sc_segbuf_nblocks -= segbuf->sb_rest_blocks;
 		ret = nilfs_sufile_free(sufile, segbuf->sb_nextnum);
 		BUG_ON(ret);
 		nilfs_segbuf_free(segbuf);
@@ -1813,8 +1837,7 @@ static int nilfs_segctor_collect(struct nilfs_sc_info *sci,
 			goto failed;
 
 		/* The current segment is filled up */
-		if (mode == SC_LSEG_DSYNC ||
-		    sci->sc_stage.main < SC_MAIN_CPFILE)
+		if (mode != SC_LSEG_SR || sci->sc_stage.main < SC_MAIN_CPFILE)
 			break;
 
 		nilfs_segctor_cancel_free_segments(sci, nilfs->ns_sufile);
@@ -1835,26 +1858,6 @@ static int nilfs_segctor_collect(struct nilfs_sc_info *sci,
 	return err;
 }
 
-/**
- * nilfs_follow_up_check - Check whether the segment is empty or not.
- * @sci: nilfs_sc_info
- *
- * We reject empty or SR-only segment if the previous write was continuing.
- */
-static int nilfs_segctor_follow_up_check(struct nilfs_sc_info *sci)
-{
-	struct nilfs_segment_buffer *segbuf = sci->sc_curseg;
-	int has_sr = (sci->sc_super_root != NULL);
-
-	if (NILFS_SEG_SIMPLEX(&segbuf->sb_sum) &&
-	    /* # of payload blocks */
-	    segbuf->sb_sum.nblocks - segbuf->sb_sum.nsumblk <= has_sr) {
-		seg_debug(2, "Aborted construction (no blocks collected)\n");
-		return 1;
-	}
-	return 0;
-}
-
 static void nilfs_list_replace_buffer(struct buffer_head *old_bh,
 				      struct buffer_head *new_bh)
 {
@@ -1990,21 +1993,39 @@ nilfs_copy_replace_page_buffers(struct page *page, struct list_head *out)
 
 	nilfs_page_add_to_lru(clone_page, 1);
 	nilfs_set_page_writeback(clone_page);
+	unlock_page(clone_page);
 	page_cache_release(clone_page);
 
 	return 0;
 }
 
+static int nilfs_test_page_to_be_frozen(struct page *page)
+{
+	struct address_space *mapping = page->mapping;
+
+	if (!mapping || !mapping->host || S_ISDIR(mapping->host->i_mode))
+		return 0;
+
+	if (page_mapped(page)) {
+		ClearPageChecked(page);
+		return 1;
+	}
+	return PageChecked(page);
+}
+
 static int nilfs_begin_page_io(struct page *page, struct list_head *out)
 {
 	if (!page || PageWriteback(page))
 		/* For split b-tree node pages, this function may be called
 		   twice.  We ignore the 2nd or later calls by this check. */
 		return 0;
-	lock_page(page);
-	nilfs_set_page_writeback(page);
 
-	if (nilfs_page_to_be_frozen(page)) {
+	lock_page(page);
+	nilfs_clear_page_dirty_for_io(page);
+	nilfs_set_page_writeback(page);
+	unlock_page(page);
+
+	if (nilfs_test_page_to_be_frozen(page)) {
 		int err = nilfs_copy_replace_page_buffers(page, out);
 		if (unlikely(err))
 			return err;
@@ -2029,7 +2050,9 @@ static int nilfs_segctor_prepare_write(struct nilfs_sc_info *sci,
 			if (bh->b_page != bd_page) {
 				if (bd_page) {
 					lock_page(bd_page);
+					clear_page_dirty_for_io(bd_page);
 					set_page_writeback(bd_page);
+					unlock_page(bd_page);
 				}
 				bd_page = bh->b_page;
 			}
@@ -2040,7 +2063,9 @@ static int nilfs_segctor_prepare_write(struct nilfs_sc_info *sci,
 			if (bh == sci->sc_super_root) {
 				if (bh->b_page != bd_page) {
 					lock_page(bd_page);
+					clear_page_dirty_for_io(bd_page);
 					set_page_writeback(bd_page);
+					unlock_page(bd_page);
 					bd_page = bh->b_page;
 				}
 				break;
@@ -2057,7 +2082,9 @@ static int nilfs_segctor_prepare_write(struct nilfs_sc_info *sci,
 	}
 	if (bd_page) {
 		lock_page(bd_page);
+		clear_page_dirty_for_io(bd_page);
 		set_page_writeback(bd_page);
+		unlock_page(bd_page);
 	}
 	err = nilfs_begin_page_io(fs_page, list);
 	if (unlikely(err))
@@ -2102,13 +2129,25 @@ static int nilfs_page_has_uncleared_buffer(struct page *page)
 	return 0;
 }
 
+static void __nilfs_end_page_io(struct page *page, int err)
+{
+	/* BUG_ON(err > 0); */
+	if (!err) {
+		if (!nilfs_page_buffers_clean(page))
+			nilfs_redirty_page(page);
+		ClearPageError(page);
+	} else {
+		nilfs_redirty_page(page);
+		SetPageError(page);
+	}
+	nilfs_end_page_writeback(page);
+}
 
 static void nilfs_end_page_io(struct page *page, int err)
 {
-	int bits;
-
 	if (!page)
 		return;
+
 	if (buffer_nilfs_node(page_buffers(page)) &&
 	    nilfs_page_has_uncleared_buffer(page))
 		/* For b-tree node pages, this function may be called twice
@@ -2117,16 +2156,7 @@ static void nilfs_end_page_io(struct page *page, int err)
 		   buffers in a split btnode page. */
 		return;
 
-	if (err < 0)
-		SetPageError(page);
-	else if (!err) {
-		bits = nilfs_page_buffers_clean(page);
-		if (bits != 0)
-			nilfs_clear_page_dirty(page, bits);
-		ClearPageError(page);
-	}
-	unlock_page(page);
-	nilfs_end_page_writeback(page);
+	__nilfs_end_page_io(page, err);
 }
 
 static void nilfs_clear_copied_buffers(struct list_head *list, int err)
@@ -2152,17 +2182,7 @@ static void nilfs_clear_copied_buffers(struct list_head *list, int err)
 			}
 		} while ((bh = bh->b_this_page) != head);
 
-		if (!err) {
-			if (nilfs_page_buffers_clean(page)) {
-				nilfs_clear_page_to_be_frozen(page);
-				__nilfs_clear_page_dirty(page);
-			}
-			ClearPageError(page);
-		} else if (err < 0)
-			SetPageError(page);
-
-		unlock_page(page);
-		end_page_writeback(page);
+		__nilfs_end_page_io(page, err);
 		page_cache_release(page);
 	}
 }
@@ -2180,10 +2200,8 @@ static void nilfs_segctor_abort_write(struct nilfs_sc_info *sci,
 		list_for_each_entry(bh, &segbuf->sb_segsum_buffers,
 				    b_assoc_buffers) {
 			if (bh->b_page != bd_page) {
-				if (bd_page) {
-					unlock_page(bd_page);
+				if (bd_page)
 					end_page_writeback(bd_page);
-				}
 				bd_page = bh->b_page;
 			}
 		}
@@ -2192,7 +2210,6 @@ static void nilfs_segctor_abort_write(struct nilfs_sc_info *sci,
 				    b_assoc_buffers) {
 			if (bh == sci->sc_super_root) {
 				if (bh->b_page != bd_page) {
-					unlock_page(bd_page);
 					end_page_writeback(bd_page);
 					bd_page = bh->b_page;
 				}
@@ -2206,10 +2223,9 @@ static void nilfs_segctor_abort_write(struct nilfs_sc_info *sci,
 			}
 		}
 	}
-	if (bd_page) {
-		unlock_page(bd_page);
+	if (bd_page)
 		end_page_writeback(bd_page);
-	}
+
 	nilfs_end_page_io(fs_page, err);
  done:
 	nilfs_clear_copied_buffers(&sci->sc_copied_buffers, err);
@@ -2257,10 +2273,8 @@ static void nilfs_segctor_complete_write(struct nilfs_sc_info *sci)
 			set_buffer_uptodate(bh);
 			clear_buffer_dirty(bh);
 			if (bh->b_page != bd_page) {
-				if (bd_page) {
-					unlock_page(bd_page);
+				if (bd_page)
 					end_page_writeback(bd_page);
-				}
 				bd_page = bh->b_page;
 			}
 		}
@@ -2282,7 +2296,6 @@ static void nilfs_segctor_complete_write(struct nilfs_sc_info *sci)
 			clear_buffer_nilfs_volatile(bh);
 			if (bh == sci->sc_super_root) {
 				if (bh->b_page != bd_page) {
-					unlock_page(bd_page);
 					end_page_writeback(bd_page);
 					bd_page = bh->b_page;
 				}
@@ -2310,10 +2323,9 @@ static void nilfs_segctor_complete_write(struct nilfs_sc_info *sci)
 	 * Since pages may continue over multiple segment buffers,
 	 * end of the last page must be checked outside of the loop.
 	 */
-	if (bd_page) {
-		unlock_page(bd_page);
+	if (bd_page)
 		end_page_writeback(bd_page);
-	}
+
 	nilfs_end_page_io(fs_page, 0);
 
 	nilfs_clear_copied_buffers(&sci->sc_copied_buffers, 0);
@@ -2566,15 +2578,17 @@ static int nilfs_segctor_do_construct(struct nilfs_sc_info *sci, int mode)
 	if (unlikely(err))
 		goto out;
 
+	if (nilfs_test_metadata_dirty(sbi))
+		set_bit(NILFS_SC_DIRTY, &sci->sc_flags);
+
+	if (nilfs_segctor_clean(sci)) {
+		seg_debug(2, "aborted (no changes found)\n");
+		goto out;
+	}
+
 	do {
 		SC_STAGE_CLEAR_HISTORY(&sci->sc_stage);
 
-		/* Re-check needs of construction */
-		if (sci->sc_stage.main == SC_MAIN_INIT &&
-		    nilfs_segctor_reconfirm(sci))
-				goto out;
-
-		/* Set next segment */
 		err = nilfs_segctor_begin_construction(sci, nilfs);
 		if (unlikely(err))
 			goto out;
@@ -2587,10 +2601,12 @@ static int nilfs_segctor_do_construct(struct nilfs_sc_info *sci, int mode)
 			goto failed;
 
 		has_sr = (sci->sc_super_root != NULL);
+
+		/* Avoid empty segment */
 		if (sci->sc_stage.main == SC_MAIN_DONE &&
-		    nilfs_segctor_follow_up_check(sci)) {
-			if (mode != SC_LSEG_DSYNC)
-				clear_bit(NILFS_SC_DIRTY, &sci->sc_flags);
+		    NILFS_SEG_EMPTY(&sci->sc_curseg->sb_sum)) {
+			seg_debug(2, "aborted (no blocks collected)\n");
+			BUG_ON(mode == SC_LSEG_SR);
 			nilfs_segctor_end_construction(sci, nilfs, 1);
 			goto out;
 		}
@@ -2642,7 +2658,8 @@ static int nilfs_segctor_do_construct(struct nilfs_sc_info *sci, int mode)
 			nilfs_segctor_commit_free_segments(sci);
 		}
 
-		nilfs_segctor_clear_metadata_dirty(sci, mode);
+		if (mode == SC_LSEG_SR)
+			nilfs_segctor_clear_metadata_dirty(sci);
 		nilfs_segctor_end_construction(sci, nilfs, 0);
 
 	} while (sci->sc_stage.main != SC_MAIN_DONE);
@@ -2693,42 +2710,33 @@ static void nilfs_segctor_start_timer(struct nilfs_sc_info *sci)
 	spin_unlock(&sci->sc_state_lock);
 }
 
-static void
-nilfs_segctor_do_flush(struct nilfs_sc_info *sci, unsigned long flag)
+static void nilfs_segctor_do_flush(struct nilfs_sc_info *sci, int bn)
 {
 	spin_lock(&sci->sc_state_lock);
-	if (!(sci->sc_state & flag)) {
-		sci->sc_state |= flag;
-		wake_up(&sci->sc_wait_daemon);
+	if (!(sci->sc_flush_request & (1 << bn))) {
+		unsigned long prev_req = sci->sc_flush_request;
+
+		sci->sc_flush_request |= (1 << bn);
+		if (!prev_req)
+			wake_up(&sci->sc_wait_daemon);
 	}
 	spin_unlock(&sci->sc_state_lock);
 }
 
 /**
  * nilfs_flush_segment - trigger a segment construction for resource control
- * @sbi: nilfs_sb_info
+ * @sb: super block
  * @ino: inode number of the file to be flushed out.
  */
-void nilfs_flush_segment(struct nilfs_sb_info *sbi, ino_t ino)
+void nilfs_flush_segment(struct super_block *sb, ino_t ino)
 {
+	struct nilfs_sb_info *sbi = NILFS_SB(sb);
 	struct nilfs_sc_info *sci = NILFS_SC(sbi);
-	unsigned long flag;
 
-	if (!sci) {
-		nilfs_warning(sbi->s_super, __func__,
-			      "Tried to flush destructed FS.\n");
-		dump_stack();
+	if (!sci || nilfs_doing_construction())
 		return;
-	}
-	if (ino >= sbi->s_nilfs->ns_first_ino)
-		flag = NILFS_SEGCTOR_FLUSH_DATA;
-	else if (ino == NILFS_IFILE_INO)
-		flag = NILFS_SEGCTOR_FLUSH_IFILE;
-	else
-		return;
-
-	seg_debug(2, "kick segment constructor (inode number=%lu)\n", ino);
-	nilfs_segctor_do_flush(sci, flag);
+	nilfs_segctor_do_flush(sci, NILFS_MDT_INODE(sb, ino) ? ino : 0);
+					/* assign bit 0 to data files */
 }
 
 int nilfs_segctor_add_segments_to_be_freed(struct nilfs_sc_info *sci,
@@ -2950,7 +2958,6 @@ int nilfs_construct_dsync_segment(struct super_block *sb,
 	}
 	spin_unlock(&sbi->s_inode_lock);
 	sci->sc_stage.dirty_file_ptr = ii;
-	sci->sc_seg_ctime = sbi->s_nilfs->ns_ctime;
 
 	seg_debug(2, "begin (mode=0x%x)\n", SC_LSEG_DSYNC);
 	err = nilfs_segctor_do_construct(sci, SC_LSEG_DSYNC);
@@ -2967,6 +2974,9 @@ struct nilfs_segctor_req {
 	int sb_err;  /* super block writeback failure */
 };
 
+#define FLUSH_FILE_BIT	(0x1) /* data file only */
+#define FLUSH_DAT_BIT	(1 << NILFS_DAT_INO) /* DAT only */
+
 static void nilfs_segctor_accept(struct nilfs_sc_info *sci,
 				 struct nilfs_segctor_req *req)
 {
@@ -2986,19 +2996,20 @@ static void nilfs_segctor_notify(struct nilfs_sc_info *sci,
 {
 	/* Clear requests (even when the construction failed) */
 	spin_lock(&sci->sc_state_lock);
-	if (req->mode == SC_FLUSH_DATA)
-		sci->sc_state &=
-			~(NILFS_SEGCTOR_COMMIT | NILFS_SEGCTOR_FLUSH_DATA);
-	else
-		sci->sc_state &=
-			~(NILFS_SEGCTOR_COMMIT | NILFS_SEGCTOR_FLUSH);
+
+	sci->sc_state &= ~NILFS_SEGCTOR_COMMIT;
 
 	if (req->mode == SC_LSEG_SR) {
 		seg_debug(3, "complete requests from seq=%d to seq=%d\n",
 			  sci->sc_seq_done + 1, req->seq_accepted);
 		sci->sc_seq_done = req->seq_accepted;
 		nilfs_segctor_wakeup(sci, req->sc_err ? : req->sb_err);
-	}
+		sci->sc_flush_request = 0;
+	} else if (req->mode == SC_FLUSH_FILE)
+		sci->sc_flush_request &= ~FLUSH_FILE_BIT;
+	else if (req->mode == SC_FLUSH_DAT)
+		sci->sc_flush_request &= ~FLUSH_DAT_BIT;
+
 	spin_unlock(&sci->sc_state_lock);
 }
 
@@ -3009,8 +3020,6 @@ static int nilfs_segctor_construct(struct nilfs_sc_info *sci,
 	struct the_nilfs *nilfs = sbi->s_nilfs;
 	int err = 0;
 
-	sci->sc_seg_ctime = nilfs->ns_ctime;
-
 	if (nilfs_discontinued(nilfs))
 		req->mode = SC_LSEG_SR;
 	if (!nilfs_segctor_confirm(sci)) {
@@ -3020,7 +3029,8 @@ static int nilfs_segctor_construct(struct nilfs_sc_info *sci,
 		seg_debug(2, "end (stage=%d)\n", sci->sc_stage.main);
 	}
 	if (likely(!err)) {
-		atomic_set(&nilfs->ns_ndirtyblks, 0);
+		if (req->mode != SC_FLUSH_DAT)
+			atomic_set(&nilfs->ns_ndirtyblks, 0);
 		if (test_bit(NILFS_SC_SUPER_ROOT, &sci->sc_flags) &&
 		    nilfs_discontinued(nilfs)) {
 			down_write(&nilfs->ns_sem);
@@ -3140,6 +3150,41 @@ static void nilfs_segctor_thread_construct(struct nilfs_sc_info *sci, int mode)
 	nilfs_transaction_unlock(sbi);
 }
 
+static void nilfs_segctor_do_immediate_flush(struct nilfs_sc_info *sci)
+{
+	int mode = 0;
+	int err;
+
+	spin_lock(&sci->sc_state_lock);
+	mode = (sci->sc_flush_request & FLUSH_DAT_BIT) ?
+		SC_FLUSH_DAT : SC_FLUSH_FILE;
+	spin_unlock(&sci->sc_state_lock);
+
+	if (mode) {
+		seg_debug(2, "begin (mode=0x%x)\n", mode);
+		err = nilfs_segctor_do_construct(sci, mode);
+		seg_debug(2, "end\n");
+
+		spin_lock(&sci->sc_state_lock);
+		sci->sc_flush_request &= (mode == SC_FLUSH_FILE) ?
+			~FLUSH_FILE_BIT : ~FLUSH_DAT_BIT;
+		spin_unlock(&sci->sc_state_lock);
+	}
+	clear_bit(NILFS_SC_PRIOR_FLUSH, &sci->sc_flags);
+}
+
+static int nilfs_segctor_flush_mode(struct nilfs_sc_info *sci)
+{
+	if (!test_bit(NILFS_SC_UNCLOSED, &sci->sc_flags) ||
+	    time_before(jiffies, sci->sc_lseg_stime + sci->sc_mjcp_freq)) {
+		if (!(sci->sc_flush_request & ~FLUSH_FILE_BIT))
+			return SC_FLUSH_FILE;
+		else if (!(sci->sc_flush_request & ~FLUSH_DAT_BIT))
+			return SC_FLUSH_DAT;
+	}
+	return SC_LSEG_SR;
+}
+
 /**
  * nilfs_segctor_thread - main loop of the segment constructor thread.
  * @arg: pointer to a struct nilfs_sc_info.
@@ -3180,15 +3225,10 @@ static int nilfs_segctor_thread(void *arg)
 
 		if (timeout || sci->sc_seq_request != sci->sc_seq_done)
 			mode = SC_LSEG_SR;
-		else if (!(sci->sc_state & NILFS_SEGCTOR_FLUSH))
+		else if (!sci->sc_flush_request)
 			break;
-		else if (!test_bit(NILFS_SC_UNCLOSED, &sci->sc_flags) ||
-			 time_before(jiffies,
-				     sci->sc_lseg_stime + sci->sc_mjcp_freq))
-			mode = (sci->sc_state & NILFS_SEGCTOR_FLUSH_IFILE) ?
-				SC_FLUSH_IFILE : SC_FLUSH_DATA;
 		else
-			mode = SC_LSEG_SR;
+			mode = nilfs_segctor_flush_mode(sci);
 
 		spin_unlock(&sci->sc_state_lock);
 		nilfs_segctor_thread_construct(sci, mode);
@@ -3219,7 +3259,7 @@ static int nilfs_segctor_thread(void *arg)
 
 		if (sci->sc_seq_request != sci->sc_seq_done)
 			should_sleep = 0;
-		else if (sci->sc_state & NILFS_SEGCTOR_FLUSH)
+		else if (sci->sc_flush_request)
 			should_sleep = 0;
 		else if (sci->sc_state & NILFS_SEGCTOR_COMMIT)
 			should_sleep = time_before(jiffies,
@@ -3383,7 +3423,7 @@ static void nilfs_segctor_destroy(struct nilfs_sc_info *sci)
 
 	spin_lock(&sci->sc_state_lock);
 	nilfs_segctor_kill_thread(sci);
-	flag = ((sci->sc_state & (NILFS_SEGCTOR_COMMIT | NILFS_SEGCTOR_FLUSH))
+	flag = ((sci->sc_state & NILFS_SEGCTOR_COMMIT) || sci->sc_flush_request
 		|| sci->sc_seq_request != sci->sc_seq_done);
 	spin_unlock(&sci->sc_state_lock);
 
