@@ -159,8 +159,10 @@ void nilfs_destroy_inode(struct inode *inode)
 #if NEED_OLD_INIT_ONCE_ARGS
 static void init_once(void *obj, struct kmem_cache *cachep,
 		      unsigned long flags)
-#else
+#elif NEED_OLD_INIT_ONCE_ARGS2
 static void init_once(struct kmem_cache *cachep, void *obj)
+#else
+static void init_once(void *obj)
 #endif
 {
 	struct nilfs_inode_info *ii = (struct nilfs_inode_info *)obj;
@@ -190,7 +192,7 @@ static int nilfs_init_inode_cache(void)
 #else
 					       init_once);
 #endif
-	return ((nilfs_inode_cachep == NULL) ? -ENOMEM : 0);
+	return (nilfs_inode_cachep == NULL) ? -ENOMEM : 0;
 }
 
 static inline void nilfs_destroy_inode_cache(void)
@@ -1093,6 +1095,7 @@ static int nilfs_remount(struct super_block *sb, int *flags, char *data)
 		if (err)
 			goto rw_remount_failed;
 
+		down_write(&nilfs->ns_sem);
 		nilfs_setup_super(sbi);
 		up_write(&nilfs->ns_sem);
 
@@ -1171,7 +1174,7 @@ static int nilfs_test_bdev_super(struct super_block *s, void *data)
 {
 	struct nilfs_super_data *sd = data;
 
-	return (s->s_bdev == sd->bdev);
+	return s->s_bdev == sd->bdev;
 }
 
 static int nilfs_test_bdev_super2(struct super_block *s, void *data)
@@ -1462,7 +1465,6 @@ static int __init init_nilfs_fs(void)
 {
 	int err;
 
-	nilfs_init_counters();
 	nilfs_init_debug_info();
 	nilfs_pages_init();
 
